# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module for implementing time-series counters."""
from __future__ import absolute_import

from collections import deque
from functools import partial
from six import next
from sparts.sparts import _Nameable, _Bindable, ProvidesCounters

import time


class SampleType:
    """Pass an array of these in the `types` paremeter to `sample()`"""
    COUNT = 'count'
    SUM = 'sum'
    AVG = 'avg'
    AVERAGE = 'avg'
    MAX = 'max'
    MIN = 'min'


class _BaseCounter(_Nameable, _Bindable, ProvidesCounters):
    """Base type for counter-like things"""
    suffix = 'UNDEF'

    def __init__(self, name=None):
        super(_BaseCounter, self).__init__(name)
        self._initialize()

    def _bind(self, obj):
        return self.__class__(name=self.name)

    def _initialize(self):
        raise NotImplementedError()

    def _genCounterCallbacks(self):
        """Return this counter's (name, value)"""
        yield self.name, self

    def getvalue(self):
        raise NotImplementedError()

    def add(self, value):
        raise NotImplementedError()

    def __call__(self):
        return self.getvalue()

    def __int__(self):
        return int(self.getvalue())

    def __float__(self):
        return float(self.getvalue())

    def __str__(self):
        v = self.getvalue()
        if v is None:
            return '__None__'
        return str(v)


class ValueCounter(_BaseCounter):
    """Base type for counter-like things that have a `._value`"""
    DEFAULT_VALUE = 0.0
    def _initialize(self, value=None):
        self._value = value or self.DEFAULT_VALUE

    def getvalue(self):
        return self._value


class Sum(ValueCounter):
    """A running total"""
    suffix = SampleType.SUM

    def add(self, value):
        self._value += value

    def increment(self):
        self.add(1.0)

    def incrementBy(self, value):
        self.add(value)

    def reset(self, value=0):
        self._value = value

counter = Sum

class Count(ValueCounter):
    """A running count"""
    suffix = SampleType.COUNT
    DEFAULT_VALUE = 0

    def add(self, value):
        self._value += 1

class Average(_BaseCounter):
    """A running average"""
    suffix = SampleType.AVERAGE

    def _initialize(self):
        self._total = 0.0
        self._count = 0

    def add(self, value):
        # TODO: Re-use sibling total/count counters if present
        # not sure how to do this sensibly
        self._total += value
        self._count += 1

    def getvalue(self):
        if self._count == 0:
            return None
        return self._total / self._count


class Max(ValueCounter):
    """A running maximum"""
    suffix = SampleType.MAX
    DEFAULT_VALUE = None

    def add(self, value):
        if self._value is None:
            self._value = value
        elif value > self._value:
            self._value = value


class Min(ValueCounter):
    """A running minimum"""
    suffix = SampleType.MIN
    DEFAULT_VALUE = None

    def add(self, value):
        if self._value is None:
            self._value = value
        elif value < self._value:
            self._value = value

    def getvalue(self):
        return self._value

# TODO: Percentiles!!


# Lookup for mapping SampleTypes to their respective classes
_SampleMethod = {
    SampleType.COUNT: Count,
    SampleType.SUM: Sum,
    SampleType.AVERAGE: Average,
    SampleType.MAX: Max,
    SampleType.MIN: Min,
}


class Samples(_Nameable, _Bindable, ProvidesCounters):
    """`samples` are used to generate series of counters dynamically

    This is so you can say, keep track of the average duration of some event for
    the last minute, hour, day, etc, and export these as 4 separate counters.
    """
    def __init__(self, types=None, windows=None, name=None):
        super(Samples, self).__init__(name)
        self.types = types or [SampleType.AVERAGE]
        # minutely, hourly
        self.windows = sorted(windows or [60, 3600])
        self.max_window = max(self.windows)
        self.samples = deque()
        self.dirty = True
        self._prev_counters = {}
        self._prev_time = None

    def _bind(self, obj):
        return self.__class__(types=self.types, windows=self.windows,
                              name=self.name)

    def _genCounterCallbacks(self):
        """Yield all the child counters."""
        for subcounter in self.iterkeys():
            yield subcounter, partial(self.getCounter, subcounter)

    def _now(self):
        """Defined to allow unittest overriding"""
        return time.time()

    def add(self, value):
        now = self._now()
        self.samples.append((now, value))

        # When adding samples, trim old ones.
        while now - self.max_window > self.samples[0][0]:
            self.samples.popleft()
        self.dirty = True

        # TODO: Handle "infinite" windows

    def getCounters(self):
        if self.dirty is False and self._prev_time == int(self._now()):
            return self._prev_counters

        ops = []
        for type in self.types:
            ops.append(_SampleMethod[type]())

        now = self._now()
        genwindows = iter(self.windows)
        window = next(genwindows)
        result = {}
        done = False

        def _saveCounterValues(window):
            """Re-usable helper function for setting results and continuing"""

            prefix = ''
            if self.name is not None:
                prefix = self.name + '.'

            for op in ops:
                result[prefix + op.suffix + '.' + str(window)] = \
                    op.getvalue()
            # Move to the next window
            try:
                return next(genwindows), False
            except StopIteration:
                # We exhausted all our windows
                return None, True

        for ts, value in reversed(self.samples):
            # We exceeded the current window
            while not done and now - window > ts:
                # Save counter values
                window, done = _saveCounterValues(window)

            if done:
                # TODO: "prune" any remaining samples
                break

            for op in ops:
                op.add(value)

        # We exhausted the samples before the windows
        while not done:
            window, done = _saveCounterValues(window)

        self._prev_counters = result
        self._prev_time = int(now)
        self.dirty = False
        return result

    def getCounter(self, name, default=None):
        return self.getCounters().get(name, default)

    def iterkeys(self):
        for type in self.types:
            for window in self.windows:
                yield self.name + '.' + type + '.' + str(window)
        # TODO: Infinite Windows


samples = Samples



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module that provides helpers for supporting thrift over HTTP in tornado."""
from __future__ import absolute_import

import tornado.web
from thrift.transport.TTransport import TMemoryBuffer
from thrift.protocol.TBinaryProtocol import TBinaryProtocol


class TornadoThriftHandler(tornado.web.RequestHandler):
    """A WebRequest handler that integrates HTTP with a thrift `Processor`.
    
    This handler MUST be initialized with a `processor` kwarg in an
    application config that includes it.
    """

    def initialize(self, processor):
        if hasattr(processor, 'processor'):
            processor = processor.processor
        self.processor = processor

    def post(self):
        """Thrift HTTP POST request.
        
        Translates the POST body to the thrift request, and returns the
        serialized thrift message in the response body.  Sets the approprate
        HTTP Content-Type header as well.
        """
        iprot = TBinaryProtocol(TMemoryBuffer(self.request.body))
        oprot = TBinaryProtocol(TMemoryBuffer())
        self.processor.process(iprot, oprot)
        self.set_header('Content-Type', 'application/x-thrift')
        self.write(oprot.trans.getvalue())



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module for tasks related to doing work from a queue"""
from six.moves import queue
from sparts.sparts import option
from sparts.vtask import VTask, ExecuteContext, TryLater

class QueueTask(VTask):
    """Task that calls `execute` for all work put on its `queue`"""
    MAX_ITEMS = 0
    WORKERS = 1
    max_items = option(type=int, default=lambda cls: cls.MAX_ITEMS,
                       help='Set a bounded queue length.  This may '
                            'cause unexpected deadlocks. [%(default)s]')
    workers = option(type=int, default=lambda cls: cls.WORKERS,
                     help='Number of threads to spawn to work on items from '
                          'its queue. [%(default)s]')

    def execute(self, item, context):
        """Implement this in your QueueTask subclasses"""
        raise NotImplementedError()

    def initTask(self):
        super(QueueTask, self).initTask()
        self.queue = queue.Queue(maxsize=self.max_items)
        self._shutdown_sentinel = object()

    def stop(self):
        super(QueueTask, self).stop()
        self.queue.put(self._shutdown_sentinel)

    def _runloop(self):
        while not self.service._stop:
            try:
                item = self.queue.get(timeout=1.0)
                if item is self._shutdown_sentinel:
                    self.queue.put(item)
                    break
            except queue.Empty:
                continue

            if isinstance(item, ExecuteContext):
                context = item
                item = context.item
            else:
                context = ExecuteContext(item=item)
            try:
                result = self.execute(item, context)
                if context.deferred is not None:
                    context.deferred.callback(result)
            except TryLater:
                context.attempt += 1
                self.queue.put(context)
            except Exception as ex:
                if context.deferred is not None:
                    self.unhandled = None
                    context.deferred.addErrback(self.unhandledErrback)
                    context.deferred.errback(ex)
                    if self.unhandled is not None:
                        self.unhandled.raiseException()
                else:
                    raise
            finally:
                self.queue.task_done()

    def unhandledErrback(self, error):
        self.unhandled = error
        return None



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Empty module for unittesting purposes."""



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
from __future__ import absolute_import

import time
from sparts.counters import counter, samples, SampleType
from sparts.sparts import option
from sparts.vtask import VTask, TryLater
from threading import Event


class PeriodicTask(VTask):
    """Task that executes `execute` at a specified interval
    
    You must either override the `INTERVAL` (seconds) class attribute, or
    pass a --{OPT_PREFIX}-interval in order for your task to run.
    """
    INTERVAL = None

    execute_duration_ms = samples(windows=[60, 240],
       types=[SampleType.AVG, SampleType.MAX, SampleType.MIN])
    n_iterations = counter()
    n_slow_iterations = counter()
    n_try_later = counter()

    interval = option(type=float, metavar='SECONDS',
                      default=lambda cls: cls.INTERVAL,
                      help='How often this task should run [%(default)s] (s)')

    def execute(self, context=None):
        """Override this to perform some custom action periodically."""
        self.logger.debug('execute')

    def initTask(self):
        super(PeriodicTask, self).initTask()
        assert self.interval is not None

        # Register an event that we can more smartly wait on in case shutdown
        # is requested while we would be `sleep()`ing
        self.stop_event = Event()

    def stop(self):
        self.stop_event.set()
        super(PeriodicTask, self).stop()

    def _runloop(self):
        t0 = time.time()
        while not self.service._stop:
            try:
                self.execute()
            except TryLater:
                self.n_try_later.increment()
                continue

            self.n_iterations.increment()
            self.execute_duration_ms.add((time.time() - t0) * 1000)
            to_sleep = (t0 + self.interval) - time.time()
            if to_sleep > 0:
                if self.stop_event.wait(to_sleep):
                    return
            else:
                self.n_slow_iterations.increment()

            t0 = time.time()



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module related to implementing fb303 thrift handlers"""
from __future__ import absolute_import

from .thrift import ThriftProcessorTask
from ..vfb303 import VServiceFB303Processor


class FB303ProcessorTask(ThriftProcessorTask):
    PROCESSOR = VServiceFB303Processor



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module that provides an API for executing and managing child processes."""
from __future__ import absolute_import

from sparts.counters import counter
from sparts.sparts import option

from twisted.internet.protocol import ProcessProtocol
from twisted.protocols.basic import LineReceiver
from sparts.tasks.twisted import TwistedTask

import functools
import signal
import time
import twisted.python.threadable
import twisted.internet.threads


class CommandTask(TwistedTask):
    """A task that provides a useful API for executing other commands.

    Python's Popen() can be hard to use, especially if you are executing long
    running child processes, and need to handle various stdout, stderr, and
    process exit events asynchronously.

    This particular implementation relies on Twisted's ProcessProtocol, but it
    wraps it in a way that makes it mostly opaque.
    """
    LOOPLESS = True
    OPT_PREFIX = 'cmd'

    kill_timeout = option(type=float, default=10.0,
                          help="Default shutdown kill timeout for outstanding "
                               "commands [%(default)s]")
    started = counter()
    finished = counter()

    def run(self, command, on_stdout=None, on_stderr=None, on_exit=None,
            line_buffered=True, kill_timeout=None):
        """Call this function to start a new child process running `command`.
        
        Additional callbacks, such as `on_stdout`, `on_stderr`, and `on_exit`,
        can be provided, that will receive a variety of parameters on the
        appropriate events.

        Line buffering can be disabled by passing `line_buffered`=False.

        Also, a custom `kill_timeout` (seconds) may be set that overrides the
        task default, in the event that a shutdown is received and you want
        to allow more time for the command to shut down."""
        self.logger.debug("task starting %s...", command)
        if isinstance(command, basestring):
            command = command.split(" ")

        # wrap on_exit with helper to remove registered comments
        on_exit = functools.partial(self._procExited, on_exit)

        proto = _ProcessProtocolAdapter(on_stdout, on_stderr, on_exit,
                                       line_buffered=line_buffered)

        if twisted.python.threadable.isInIOThread():
            result = self.reactor.spawnProcess(proto, executable=command[0],
                                               args=command)
        else:
            result = twisted.internet.threads.blockingCallFromThread(
                self.reactor, self.reactor.spawnProcess, proto,
                executable=command[0], args=command)

        self.outstanding[result] = kill_timeout

        self.started.increment()
        return result

    def initTask(self):
        super(CommandTask, self).initTask()
        self.outstanding = {}

    def _procExited(self, on_exit, proto, trans, reason):
        self.logger.debug("%s closed for %s", trans, reason)
        if on_exit is not None:
            on_exit(reason)

        self.outstanding.pop(trans)

        self.finished.increment()
        return None

    def join(self):
        """Overridden to block for process workers to shutdown / be killed."""
        # TODO: Conditions instead of sleep polling?
        while len(self.outstanding) > 0:
            time.sleep(0.250)

    def _killOutstanding(self, trans):
        if trans in self.outstanding:
            self.logger.info("Sending SIGKILL to %s", trans)
            trans.signalProcess(signal.SIGKILL)

    def stop(self):
        # twisted is pretty smart; the default signal handlers it installs
        # propagate SIGTERM to its children, so while we don't need to manually
        # TERM, we might still need to set some kill timeouts
        super(CommandTask, self).stop()

        for trans, kill_timeout in self.outstanding.items():
            if kill_timeout is None:
                kill_timeout = self.kill_timeout

            self.logger.info("Enqueuing kill for %s in %.1fs",
                             trans, kill_timeout)
            args = (kill_timeout, self._killOutstanding, trans)
            if twisted.python.threadable.isInIOThread():
                self.reactor.callLater(*args)
            else:
                self.reactor.callFromThread(self.reactor.callLater, *args)

    def isDoneWithReactor(self):
        """Overridden to keep reactor running until all commands finish."""
        return len(self.outstanding) == 0

class _ProcessProtocolAdapter(ProcessProtocol):
    """ProcessProtocol that allows custom callbacks, buffering."""
    def __init__(self, on_stdout=None, on_stderr=None, on_exit=None,
                 line_buffered=True):

        self.line_buffered = line_buffered
        self.on_stdout = on_stdout or (lambda *args: None)
        self.on_stderr = on_stderr or (lambda *args: None)
        self.on_exit = on_exit or (lambda *args: None)

        if self.line_buffered:
            self.stdout_buffer = self._makeBuffer(self.on_stdout)
            self.stderr_buffer = self._makeBuffer(self.on_stderr)

    def _makeBuffer(self, callback):
        buf = LineReceiver()
        buf.delimiter = "\n"
        buf.lineReceived = lambda line: callback(self.transport, line)
        return buf

    def connectionMade(self):
        ProcessProtocol.connectionMade(self)
        self.original_pid = self.transport.pid

    def outReceived(self, data):
        if self.line_buffered:
            self.stdout_buffer.dataReceived(data)
        else:
            self.on_stdout(self.transport, data)

    def errReceived(self, data):
        if self.line_buffered:
            self.stderr_buffer.dataReceived(data)
        else:
            self.on_stderr(self.transport, data)

    def processEnded(self, reason):
        # Ugh, this is a hack... I need to pass self because Twisted unsets pid
        # after the process has exited, but I need the *old* pid to do proper
        # accounting.  *YES* I get why this is fundamentally broken
        self.on_exit(self, self.transport, reason)
        self.transport.loseConnection()



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Helpers for commonly performed file operations"""
from distutils.spawn import find_executable
import errno
import logging
import os.path
import shutil
import tempfile

logger = logging.getLogger('sparts.fileutils')


def readfile(path):
    """Return the contents of the file at `path`"""
    with open(path, mode='rb') as f:
        return f.read().decode('UTF-8')

def writefile(path, contents):
    """Write `contents` to the file at `path`"""
    logger.debug('writefile("%s", ...)', path)
    with open(path, mode='wb') as f:
        return f.write(contents.encode('UTF-8'))

def makedirs(path, *args, **kwargs):
    """Create necessary directory heirarchy to `path` if it doesn't exist"""
    try:
        logger.debug('makedirs("%s", ...)', path)
        os.makedirs(path, *args, **kwargs)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise


# This function is really handy.  Make it accessible via this module
find_executable = find_executable


def resolve_partition(path):
    """Return the mount that `path` is present on.
    
    Requires `psutil`."""
    import psutil
    path = os.path.realpath(path)
    found = None

    for partition in psutil.disk_partitions('all'):
        mountpoint = partition.mountpoint
        if not mountpoint.endswith('/'):
            mountpoint += '/'
        if path.startswith(mountpoint):
            if found is None:
                found = partition
            elif len(mountpoint) > len(found.mountpoint):
                found = partition
    return found


class NamedTemporaryDirectory(object):
    """Wrapper around `mkdtemp` that cleans up after itself

    These objects provide the following additional nice features:
    
    - A `name` attribute, which refers to the full temporary path created
    - The ContextManager protocol, similar to ctx.tempdir,
      to cleanup as the context exits.
    - Implicit cleanup on dereference (via `__del__`)
    - Helpers for reading and writing files relative to this created directory
    """
    def __init__(self, suffix="", prefix=tempfile.template, dir=None): 
        self.name = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
        self.delete = True
        self.close_called = False

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc, value, tb):
        self.close()

    def __str__(self):
        return self.name

    def __repr__(self):
        """Overridden to include the temp dir path for debugging clarity."""
        return "<NamedTemporaryDirectory '%s' at 0x%x>" % (self.name, id(self))

    def close(self):
        """Trigger cleanup (if it has not been disabled)"""
        if not self.close_called:
            if self.delete:
                shutil.rmtree(self.name)
            self.close_called = True

    def keep(self):
        """Request persistence of the contents of this temporary directory.
        
        Disabled the cleanup logic that is normally triggered on `__exit__()`,
        `__del__()`, or close() calls."""
        self.delete = False

    def writefile(self, path, contents):
        """Writes `contents` to the `path` relative to this directory"""
        return writefile(self.join(path), contents)

    def readfile(self, path):
        """Reads the contents from the `path` relative to this directory"""
        return readfile(self.join(path))

    def symlink(self, path, dst):
        """Create a symlink to `dst` at the `path` relative to this directory"""
        return os.symlink(dst, self.join(path))

    def join(self, *path):
        return os.path.join(self.name, *path)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module for common base classes and helpers, such as options and counters"""
from collections import namedtuple
from functools import partial
from six import iteritems


class _Nameable(object):
    """Base class for attribute classes with automatically set `name` attribute"""
    def __init__(self, name):
        super(_Nameable, self).__init__()
        self.name = name

    def _getNameForIdentifier(self, name):
        return name


class _Bindable(object):
    """Helper class for allowing instance-unique class-declarative behavior."""
    def __init__(self, *args, **kwargs):
        self._bound = {}
        super(_Bindable, self).__init__(*args, **kwargs)

    def __get__(self, instance, owner):
        if owner is None:
            return self

        if owner not in self._bound:
            self._bound[owner] = self._bind(owner)

        return self._bound[owner]

    def _bind(self, obj):
        raise NotImplementedError()

class ProvidesCounters(object):
    def _genCounterCallbacks(self):
        """Yields this item's (names, value) counter tuple(s)."""
        raise NotImplementedError()

_AddArgArgs = namedtuple('_AddArgArgs', ['opts', 'kwargs'])

class option(_Nameable):
    def __init__(self, name=None, type=None, default=None, help=None,
                 action=None, metavar=None, required=False, choices=None,
                 nargs=None):
        super(option, self).__init__(name)

        # Set defaults for action=storeX to bool (otherwise, str)
        if type is None:
            if action in ['store_true', 'store_false']:
                type = bool
            else:
                type = str

        self.type = type
        self.default = default
        self.help = help
        self.action = action
        self.metavar = metavar
        self.required = required
        self.choices = choices
        self.nargs = nargs

    def __get__(self, obj, type=None):
        if obj is None:
            return self

        value = self._getter(obj)(self.name)
        if self.type is not None:
            value = self.type(value)
        return value

    def __set__(self, obj, value):
        if self.type is not None:
            value = self.type(value)
        self._setter(obj)(self.name, value)

    def _getter(self, obj):
        getter = getattr(obj, 'getTaskOption', None)
        if getter is None:
            getter = getattr(obj, 'getOption', None)
        assert getter is not None
        return getter

    def _setter(self, obj):
        setter = getattr(obj, 'setTaskOption', None)
        if setter is None:
            setter = getattr(obj, 'setOption', None)
        assert setter is not None
        return setter

    def _prepareForArgumentParser(self, task_cls):
        """Convert inst properties to *args, **kwargs for ap.add_argument().
        """
        name = task_cls._loptName(self.name)

        # This is kinda funky.  I want to support some form of shorthand
        # notation for overridable default values.  Doing it this way means we
        # can do option(..., default=lambda cls: cls.FOO, ...)
        default = self.default
        if callable(default):
            default = default(task_cls)

        kwargs = dict(default=default, help=self.help, action=self.action,
                      required=self.required)
        if self.action is None:
            kwargs['metavar'] = self.metavar
            kwargs['type'] = self.type
            kwargs['choices'] = self.choices
        if self.nargs is not None:
            kwargs['nargs'] = self.nargs
        return _AddArgArgs([name], kwargs)

    def _addToArgumentParser(self, optargs, ap):
        ap.add_argument(*optargs.opts, **optargs.kwargs)

    def _getNameForIdentifier(self, name):
        return name.replace('_', '-')


class _NameHelper(type):
    def __new__(cls, name, bases, attrs):

        for k, v in iteritems(attrs):
            # Assign `name` for options
            if not isinstance(v, _Nameable):
                continue
            if v.name is not None:
                continue
            v.name = v._getNameForIdentifier(k)
        return super(_NameHelper, cls).__new__(cls, name, bases, attrs)


_SpartsObjectBase = _NameHelper('_SpartsObjectBase', (object, ), {})

class _SpartsObject(_SpartsObjectBase):
    def __new__(cls, *args, **kwargs):
        inst = super(_SpartsObject, cls).__new__(cls)
        inst.counters = {}
        #for k, v in iteritems(cls.__dict__):
        for k in dir(cls):
            v = getattr(cls, k)
            if isinstance(v, ProvidesCounters):
                for cn, cv in v._genCounterCallbacks():
                    inst.counters[cn] = cv

        return inst

    @classmethod
    def _loptName(cls, name):
        raise NotImplementedError()

    def getCounters(self):
        result = {}
        for k in self.counters:
            result[k] = self.getCounter(k)

        for cn, c in iteritems(self.getChildren()):
            for k in c.counters:
                result[cn + '.' + k] = c.getCounter(k)

        return result

    def getCounter(self, name):
        return self.counters.get(name, lambda: None)

    def getChildren(self):
        return {}

    @classmethod
    def _addArguments(cls, ap):
        options = get_options(cls)
        for opt in options:
            opt.regfunc(ap)


_OptRegFunc = namedtuple('_OptRegFunc', ['opt', 'regfunc'])

def get_options(cls):
    """Get argparse options for class, `cls`

    Look for class level attributes of type option and
    convert them into the arguments  necessary for calling
    parser.add_argument().

    Arguments:
        subclass of VTask or Vservice - `cls`

    Returns:
        list(namedtuple) -
            .opt - namedtuple
                   .args - list of argument string names
                          (e.g. ['--help', '-h'])
                   .kwargs - dict of kwargs for add_argument()
                          (e.g. default=False, action='store_true' etc)
            .regfunc - callable that takes the ArgumentParser as an argument
                       and adds the option to it.
                       (e.g. "foo.regfunc(ap)" registers the foo option on ap)
    """
    ret = []
    for k in dir(cls):
        v = getattr(cls, k)
        preparefunc = getattr(v, '_prepareForArgumentParser', None)
        if not preparefunc:
            continue
        opt = preparefunc(cls)
        regfunc = partial(getattr(v, '_addToArgumentParser'), opt)
        ret.append(_OptRegFunc(opt, regfunc))
    return ret



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module containing various helpful context managers."""
from __future__ import absolute_import

from contextlib import contextmanager
from sparts.fileutils import NamedTemporaryDirectory

import copy
import os.path
import sys

@contextmanager
def tmpdir(*args, **kwargs):
    """Create a NamedTemporaryDirectory "as" the str path"""
    with NamedTemporaryDirectory() as d:
        yield d.name

@contextmanager
def add_path(path, index=None):
    """Temporarily add `path` to the PYTHONPATH. Not thread-safe.
    
    If `index` is None, append to the end, otherwise, use `index` as specified
    to `list.insert()`"""
    if index is None:
        sys.path.append(path)
    else:
        sys.path.insert(index, path)
    try:
        yield path
    finally:
        sys.path.remove(path)

@contextmanager
def chdir(path):
    """Temporarily chdir to `path`.  Not thread-safe."""
    assert os.path.exists(path)
    assert os.path.isdir(path)
    old_path = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_path)

@contextmanager
def module_snapshot():
    """Removes any new modules from the module cache after context exit"""
    try:
        modules = copy.copy(sys.modules)
        yield
    finally:
        to_delete = [k for k in sys.modules if k not in modules]
        for k in to_delete:
            del sys.modules[k]



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module containing helpers for implementing fb303 with `VService`s"""
from .fb303 import FacebookService
from .fb303.ttypes import fb_status


class VServiceFB303Processor(FacebookService.Processor):
    def __init__(self, service):
        FacebookService.Processor.__init__(self, self)
        self.service = service

    def getName(self):
        return self.service.name

    def getVersion(self):
        return str(self.service.VERSION)

    def getStatus(self):
        # TODO: DEAD?  STARTING?  STOPPED?
        if self.service._stop:
            return fb_status.STOPPING
        for task in self.service.tasks:
            # Only check LOOPLESS tasks for "dead" threads
            if not task.LOOPLESS:
                for thread in task.threads:
                    if not thread.isAlive():
                        return fb_status.WARNING

        # Return WARNING if there are any registered warnings
        if self.service.getWarnings():
            return fb_status.WARNING
        return fb_status.ALIVE

    def getStatusDetails(self):
        messages = []
        if self.service._stop:
            messages.append('%s is shutting down' % (self.getName()))

        # Check for dead threads
        for task in self.service.tasks:
            if not task.LOOPLESS:
                for thread in task.threads:
                    if not thread.isAlive():
                        messages.append('%s has dead threads!' % task.name)

        # Append any registered warnings
        messages.extend(self.service.getWarnings().values())
        return '\n'.join(messages)

    def getCounters(self):
        result = {}
        for k, v in self.service.getCounters().iteritems():
            v = v()
            if v is None:
                continue
            result[k] = int(v)
        return result

    def getCounter(self, name):
        result = self.service.getCounter(name)()
        if result is None:
            raise ValueError("%s is None" % (name))
        return int(result)

    def setOption(self, name, value):
        if value == '__None__':
            value = None
        else:
            cur_value = getattr(self.service.options, name)
            if cur_value is not None:
                try:
                    value = cur_value.__class__(value)
                except Exception as e:
                    self.logger.debug('Unable to cast %s to %s (%s)', value,
                                      cur_value.__class__, e)
        self.service.setOption(name, value)

    def getOption(self, name):
        value = self.service.getOption(name)
        if value is None:
            value = '__None__'
        return value

    def getOptions(self):
        result = {}
        for k in self.service.getOptions():
            result[k] = self.getOption(k)
        return result

    def aliveSince(self):
        return self.service.start_time

    def reinitialize(self):
        self.service.restart()

    def shutdown(self):
        self.service.shutdown()



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""This module contains tornado-related helper tasks and classes."""
from __future__ import absolute_import

from six import itervalues
from sparts.counters import counter  #, samples, SampleType
from sparts.sparts import option
from sparts.vtask import VTask, SkipTask

import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.netutil

import grp
import os


class TornadoIOLoopTask(VTask):
    """Configure and run the Tornado IO Loop in a sparts task"""
    OPT_PREFIX = 'tornado'

    def initTask(self):
        super(TornadoIOLoopTask, self).initTask()
        needed = getattr(self.service, 'REQUIRE_TORNADO', False)
        for t in self.service.tasks:
            if isinstance(t, TornadoTask):
                needed = True

        if not needed:
            raise SkipTask("No TornadoTasks found or enabled")

        self.ioloop = tornado.ioloop.IOLoop.instance()

    def _runloop(self):
        self.ioloop.start()

    def stop(self):
        self.ioloop.stop()
        super(TornadoIOLoopTask, self).stop()


class TornadoTask(VTask):
    """Base class for tasks that require the tornado IO Loop.
    
    Implicitly configures the tornado IO Loop task as a dependency.
    
    The ioloop can be accessed via `self.ioloop`"""
    DEPS = [TornadoIOLoopTask]

    def initTask(self):
        super(TornadoTask, self).initTask()
        self.ioloop_task = self.service.requireTask('TornadoIOLoopTask')

    @property
    def ioloop(self):
        return self.ioloop_task.ioloop


class TornadoHTTPTask(TornadoTask):
    """A loopless task that implements an HTTP server using Tornado.
    
    It is loopless because it depends on tornado's separate IOLoop task.  You
    will need to subclass this to do something more useful."""
    LOOPLESS = True
    OPT_PREFIX = 'http'
    DEFAULT_PORT = 0
    DEFAULT_HOST = ''
    DEFAULT_SOCK = ''

    requests = counter()
    #latency = samples(windows=[60, 3600],
    #                  types=[SampleType.AVG, SampleType.MIN, SampleType.MAX])

    host = option(metavar='HOST', default=lambda cls: cls.DEFAULT_HOST,
                  help='Address to bind server to [%(default)s]')
    port = option(metavar='PORT', default=lambda cls: cls.DEFAULT_PORT,
                  help='Port to run server on [%(default)s]')
    sock = option(metavar='PATH', default=lambda cls: cls.DEFAULT_SOCK,
                  help='Default path to use for local file socket '
                       '[%(default)s]')
    group = option(name='sock-group', metavar='GROUP', default='',
                   help='Group to create unix files as [%(default)s]')

    def getApplicationConfig(self):
        """Override this to register custom handlers / routes."""
        return [
            ('/', HelloWorldHandler),
        ]

    def initTask(self):
        super(TornadoHTTPTask, self).initTask()

        self.app = tornado.web.Application(
            self.getApplicationConfig(),
            log_function=self.tornadoRequestLog)

        self.server = tornado.httpserver.HTTPServer(self.app)

        if self.sock:
            assert self.host == self.DEFAULT_HOST, \
                "Do not specify host *and* sock (%s, %s)" % \
                (self.host, self.sock)
            assert int(self.port) == self.DEFAULT_PORT, \
                "Do not specify port *and* sock (%s, %s)" % \
                (self.port, self.DEFAULT_PORT)

            gid, mode = -1, 0o600
            if self.group != '':
                e = grp.getgrnam(self.group)
                gid, mode = e.gr_gid, 0o660

            sock = tornado.netutil.bind_unix_socket(self.sock, mode=mode)
            if gid != -1:
                os.chown(self.sock, -1, gid)
            self.server.add_sockets([sock])
        else:
            self.server.listen(self.port, self.host)

        self.bound_addrs = []
        for sock in itervalues(self.server._sockets):
            sockaddr = sock.getsockname()
            self.bound_addrs.append(sockaddr)
            self.logger.info("%s Server Started on %s (port %s)",
                             self.name, sockaddr[0], sockaddr[1])

    @property
    def bound_v4_addrs(self):
        return [a[0] for a in self.bound_addrs if len(a) == 2]

    @property
    def bound_v6_addrs(self):
        return [a[0] for a in self.bound_addrs if len(a) == 4]

    def tornadoRequestLog(self, handler):
        self.requests.increment()

    def stop(self):
        super(TornadoHTTPTask, self).stop()
        self.server.stop()


class HelloWorldHandler(tornado.web.RequestHandler):
    """A sample twisted web handler for use in the default http server task"""
    def get(self):
        self.write("Hello, world")



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module providing tasks that help with dbus integration"""
from __future__ import absolute_import

from ..vtask import VTask, SkipTask
from ..sparts import option
from ..fb303.dbus import FB303DbusService
from .fb303 import FB303ProcessorTask

from dbus.mainloop.glib import DBusGMainLoop
import dbus
import dbus.service
import gobject
import glib
import time


class VServiceDBusObject(dbus.service.Object):
    """DBus interface implementation that exports common VService methods"""
    def __init__(self, dbus_service):
        self.dbus_service = dbus_service
        self.service = self.dbus_service.service
        self.logger = self.dbus_service.logger
        self.path = '/'.join(['', self.service.name, 'sparts'])
        dbus.service.Object.__init__(self, self.dbus_service.bus, self.path)

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='s', out_signature='v')
    def getOption(self, name):
        return self.service.getOption(name)

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='sv', out_signature='')
    def setOption(self, name, value):
        if value == '__None__':
            value = None
        self.service.setOption(name, value)

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='', out_signature='a{sv}')
    def getOptions(self):
        result = {}
        for k, v in self.service.getOptions().iteritems():
            # dbus doesn't support serializing None as a variant
            if v is None:
                v = '__None__'
            result[k] = v
        return result

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='', out_signature='as')
    def listOptions(self):
        return self.service.getOptions().keys()

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='', out_signature='')
    def shutdown(self):
        self.service.shutdown()

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='', out_signature='')
    def restart(self):
        self.service.reinitialize()

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='', out_signature='as')
    def listTasks(self):
        return [t.name for t in self.service.tasks]

    @dbus.service.method(dbus_interface='org.sparts.Service',
                         in_signature='', out_signature='x')
    def uptime(self):
        return int(time.time() - self.service.start_time)


class DBusMainLoopTask(VTask):
    """Configure and run the DBus Main Loop in a sparts task"""
    THREADS_INITED = False
    mainloop = None

    def initTask(self):
        super(DBusMainLoopTask, self).initTask()
        needed = getattr(self.service, 'REQUIRE_DBUS', False)
        for t in self.service.tasks:
            if isinstance(t, DBusTask):
                needed = True

        if not needed:
            raise SkipTask("No DBusTasks found or enabled")

        self.dbus_loop = DBusGMainLoop(set_as_default=True)
        glib.threads_init()
        gobject.threads_init()
        dbus.mainloop.glib.threads_init()
        self.mainloop = gobject.MainLoop()

    def _runloop(self):
        self.mainloop.run()

    def stop(self):
        super(DBusMainLoopTask, self).stop()

        if self.mainloop is None:
            return

        self.mainloop.quit()

        # OK!  Apparently, there is some wonky destructor event handling that
        # seems to work better than just calling .quit() in order to properly
        # return full control of signal handling, threads, etc to the actual
        # main process.
        self.mainloop = None

class DBusTask(VTask):
    """Base Class for Tasks that depend on the DBus Main Loop"""
    DEPS = [DBusMainLoopTask]
    LOOPLESS = True

    def initTask(self):
        super(DBusTask, self).initTask()
        self.mainloop_task = self.service.requireTask(DBusMainLoopTask)

    @property
    def mainloop(self):
        return self.mainloop_task.mainloop


class DBusServiceTask(DBusTask):
    """Glue Task for exporting this VService over dbus"""
    OPT_PREFIX = 'dbus'
    BUS_NAME = None
    BUS_CLASS = VServiceDBusObject

    bus_name = option(default=lambda cls: cls.BUS_NAME, metavar='NAME',
                      help='Bus Name.  Should be something like '
                           '"com.sparts.AwesomeService"')
    replace = option(action='store_true', type=bool,
        default=False, help='Replace, and enable replacing of this service')
    queue = option(action='store_true', type=bool,
        default=False, help='If not --{task}-replace, will wait to take '
                            'this bus name')

    def initTask(self):
        super(DBusServiceTask, self).initTask()

        assert self.bus_name is not None, \
            "You must pass a --{task}-bus-name"

    def start(self):
        self.bus = dbus.SessionBus(private=True)
        self.dbus_service = dbus.service.BusName(self.bus_name, self.bus,
            self.replace, self.replace, self.queue)
        self.addHandlers()
        super(DBusServiceTask, self).start()

    def addHandlers(self):
        self.sparts_dbus = self.BUS_CLASS(self)
        task = self.service.getTask(FB303ProcessorTask)
        if task is not None:
            self.fb303_dbus = FB303DbusService(
                self.dbus_service, task.processor, self.service.name)

    def stop(self):
        del(self.dbus_service)
        #self.bus.close()
        super(DBusServiceTask, self).stop()



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Base Task and related helper classes for sparts' task system

Tasks in sparts are a way to organize and delegate some sort of
background or other synchronized processing.  This module defines
the most common features.
"""
from __future__ import absolute_import
import logging
import threading
from six.moves import xrange
from sparts.sparts import _SpartsObject


_REGISTERED_TASKS = set()

class VTask(_SpartsObject):
    """The base class for all tasks.  Needs to be subclassed to be useful.

    Attributes:
        OPT_PREFIX - Overrides the prefix for any associated options
        LOOPLESS - True indicates this task should not spawn any threads
        DEPS - List of `VTask` subclasses that must be initialized first
        workers - Number of Threads that should execute the `_runloop`

    """

    OPT_PREFIX = None
    LOOPLESS = False
    DEPS = []
    workers = 1

    @property
    def name(self):
        return self.__class__.__name__

    def __init__(self, service):
        """Task Constructor.  requires a `service` VService instance

        You should not need to override this.  Override initTask isntead."""
        self.service = service
        self.logger = logging.getLogger('%s.%s' % (service.name, self.name))
        self.threads = []

    def initTask(self):
        """Override this to do any task-specific initialization

        Don't forget to call super(...).initTask(), or things may not
        run properly."""
        if not self.LOOPLESS:
            for i in xrange(self.workers):
                if self.workers == 1:
                    name = self.name
                else:
                    name = '%s-%d' % (self.name, i + 1)
                self.threads.append(
                    threading.Thread(target=self._run, name=name))

    def initTaskThread(self):
        """Override thread-specific initialization for multi-threaded tasks"""

    def start(self):
        """Called during bootstrap to spin up threads post-creation."""
        if not self.LOOPLESS:
            for thread in self.threads:
                thread.start()

    def stop(self):
        """Custom stopping logic for this task.

        This is called by the main VService thread, after a graceful shutdown
        request has been received."""
        pass

    def join(self):
        """Block, waiting for all child worker threads to finish."""
        if not self.LOOPLESS:
            for thread in self.threads:
                while thread.isAlive():
                    thread.join(0.5)

    @property
    def running(self):
        """Returns True if task is still doing work.

        This base implementation returns True if any child threads are alive"""
        for thread in self.threads:
            if thread.isAlive():
                return True
        return False

    def _run(self):
        try:
            self.initTaskThread()
            self._runloop()
        except Exception:
            # In general, you should not get here.  So, we will shutdown the
            # server.  It is better for your service to *completely* crash in
            # response to an unhandled error, than to continue on in some sort
            # of half-alive zombie state.  Please catch your exceptions.
            # Consider throwing a TryLater if this task is a subclass of
            # QueueTask or PeriodicTask.
            #
            # I hate zombies.
            self.logger.exception("Unhandled exception in %s", self.name)
            self.service.shutdown()
        finally:
            self.logger.debug('Thread %s exited',
                              threading.currentThread().name)

    def _runloop(self):
        """For normal (non-LOOPLESS) tasks, this MUST be implemented"""
        # TODO: May require some janky metaprogramming to make ABC enforce
        # this in a cleaner way.
        raise NotImplementedError()

    @classmethod
    def _loptName(cls, name):
        return '--' + cls._optName(name).replace('_', '-')

    @classmethod
    def _optName(cls, name):
        parts = [cls.OPT_PREFIX or cls.__name__,
                 name.replace('-', '_')]
        return '_'.join(parts)

    def getTaskOption(self, opt, default=None):
        return getattr(self.service.options,
                       self._optName(opt), default)

    @classmethod
    def register(cls):
        _REGISTERED_TASKS.add(cls)


def get_registered_tasks():
    return _REGISTERED_TASKS.copy()


class SkipTask(Exception):
    """Throw during initTask() to skip execution of this task.

    Useful in case the task is missing configuration critical to its operation,
    but not critical to the overall program.

    A good example might be a network-based logging task."""
    pass


class TryLater(Exception):
    """Throw this in overridden tasks to defer execution.

    Can be used to temporarily suspend and restart execution, which is useful
    for handling unexpected error conditions, or re-scheduling work."""
    pass


class ExecuteContext(object):
    """An abstraction used internally by various tasks to track work

    Encapsulates common metrics for work that can be retried later, hooks for
    signalling completion, etc"""
    def __init__(self, attempt=1, item=None, deferred=None):
        self.attempt = attempt
        self.item = item
        self.deferred = deferred


def resolve_dependencies(task_classes):
    """Returns a flattened dependency chain for `task_classes`.

    Turns a short list of `Task` subclasses into a longer list of `Task`
    subclasses.  This is useful for tasks that depend on other tasks, such as
    TwistedTask subclasses that must initialize and run after the Twisted
    reactor has been started."""
    result = []
    for t in task_classes:
        assert issubclass(t, VTask)
        for dep in resolve_dependencies(t.DEPS):
            if dep not in result:
                result.append(dep)

        if t not in result:
            result.append(t)
    return result



#
# Autogenerated by Thrift Compiler (1.0.0-dev)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class nagios_result(object):
  OK = 0
  WARNING = 1
  CRITICAL = 2
  UNKNOWN = 3

  _VALUES_TO_NAMES = {
    0: "OK",
    1: "WARNING",
    2: "CRITICAL",
    3: "UNKNOWN",
  }

  _NAMES_TO_VALUES = {
    "OK": 0,
    "WARNING": 1,
    "CRITICAL": 2,
    "UNKNOWN": 3,
  }




# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
from __future__ import absolute_import

import dbus.service
from .ttypes import fb_status

import datetime
import logging
import functools
import os


# TODO- move this somewhere else
def log_unhandled(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            logging.getLogger('sparts.dbus.fb303') \
                .exception('Unhandled Exception in dbus method')
            raise
    return wrapped

class FB303DbusService(dbus.service.Object):
    def __init__(self, bus, handler, name=None):
        parts = ['', 'fb303']
        if name:
            parts.insert(1, name)
        path = '/'.join(parts)
        dbus.service.Object.__init__(self, bus, path)
        self.bus = bus
        self.handler = handler
        self.logger = logging.getLogger('sparts.fb303.dbus')

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='s')
    def getName(self):
        return self.handler.getName()

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='s')
    def getVersion(self):
        return self.handler.getVersion()

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='s')
    def getStatus(self):
        return fb_status._VALUES_TO_NAMES[self.handler.getStatus()]

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='s')
    def getStatusDetails(self):
        return self.handler.getStatusDetails()

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='a{sx}')
    def getCounters(self):
        return self.handler.getCounters()

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='s', out_signature='x')
    def getCounter(self, key):
        return self.handler.getCounter(key)

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='ss', out_signature='')
    def setOption(self, key, value):
        if value == '__None__':
            value = None
        return self.handler.setOption(key, value)

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='s', out_signature='s')
    def getOption(self, key):
        value = self.handler.getOption(key)
        if value is None:
            value = '__None__'
        return str(value)

    @log_unhandled
    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='a{ss}')
    def getOptions(self):
        result = {}
        for k in self.handler.getOptions().iterkeys():
            result[k] = self.getOption(k)
        return result

    @log_unhandled
    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='i', out_signature='s')
    def getCpuProfile(self, profileDurationInSec):
        logging.getLogger('sparts.dbus').debug('[%s] startProfile', os.getpid())
        return self.handler.getCpuProfile(profileDurationInSec)

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='x')
    def aliveSince(self):
        return self.handler.aliveSince()

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='')
    def reinitialize(self):
        return self.handler.reinitialize()

    @dbus.service.method(dbus_interface='com.facebook.fb303.Service',
                         in_signature='', out_signature='')
    def shutdown(self):
        return self.handler.shutdown()

    @dbus.service.method(dbus_interface='org.sparts.FB303Service',
                         in_signature='sv', out_signature='')
    def setOptionV(self, key, value):
        self.handler.setOption(key, str(value))


    @dbus.service.method(dbus_interface='org.sparts.FB303Service',
                         in_signature='', out_signature='s')
    def aliveSinceStr(self):
        return datetime.datetime.fromtimestamp(self.handler.aliveSince()) \
                .strftime('%Y-%m-%d %H:%M:%S')



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""vservice defines the base service class, `VService`

VService can be used directly, for example with `VService.initFromCLI()`,
or it can be subclassed and used similarly.
"""
from __future__ import absolute_import
from __future__ import print_function

import logging
import signal
import sys
import threading
import time

from argparse import ArgumentParser
from .compat import OrderedDict

from .vtask import SkipTask, resolve_dependencies, get_registered_tasks
from .deps import HAS_PSUTIL
from .sparts import _SpartsObject, option


class VService(_SpartsObject):
    """Core class for implementing services."""
    DEFAULT_LOGLEVEL = 'DEBUG'
    REGISTER_SIGNAL_HANDLERS = True
    TASKS = []
    VERSION = ''
    _name = None
    dryrun = option(action='store_true', help='Run in "dryrun" mode')
    level = option(default=DEFAULT_LOGLEVEL, help='Log Level [%(default)s]')
    register_tasks = option(name='tasks', default=None,
                            metavar='TASK', nargs='*',
                            help='Tasks to run.  Pass without args to see the '
                                 'list. If not passed, all tasks will be '
                                 'started')
    if HAS_PSUTIL:
        runit_install = option(action='store_true',
                               help='Install this service under runit.')

    def __init__(self, ns):
        super(VService, self).__init__()
        self.logger = logging.getLogger(self.name)
        self.options = ns
        self.initLogging()
        self._stop = False
        self._restart = False
        self.tasks = []
        self.warnings = OrderedDict()
        self.warning_id = 0
        self.start_time = time.time()

    def initService(self):
        """Override this to do any service-specific initialization"""

    @classmethod
    def _resolveDependencies(cls):
        tasks = set(cls.TASKS).union(get_registered_tasks())
        return resolve_dependencies(tasks)

    @classmethod
    def _loptName(cls, name):
        return '--' + name.replace('_', '-')

    def preprocessOptions(self):
        """Processes "action" oriented options."""
        if self.getOption('runit_install'):
            self._install()

        if self.options.tasks == []:
            print("Available Tasks:")
            for t in self._resolveDependencies():
                print(" - %s" % t.__name__)
            sys.exit(1)

    def _createTasks(self):
        all_tasks = self._resolveDependencies()

        selected_tasks = self.options.tasks
        if selected_tasks is None:
            selected_tasks = [t.__name__ for t in all_tasks]

        for t in all_tasks:
            if t.__name__ in selected_tasks:
                self.tasks.append(t(self))

        # Call service initialization hook after tasks have been instantiated,
        # but before they've been initialized.
        self.initService()

        exceptions = []
        required = []
        for t in self.tasks:
            try:
                t.initTask()
                required.append(t)
            except SkipTask as e:
                self.logger.info("Skipping %s (%s)", t.name, e)
            except Exception as e:
                self.logger.exception("Error creating task, %s", t.name)
                exceptions.append(e)
        self.tasks = required

        if len(exceptions):
            raise Exception("Unable to start service (%d task start errors)" %
                            len(exceptions))

    def _handleShutdownSignals(self, signum, frame):
        assert signum in (signal.SIGINT, signal.SIGTERM)
        self.logger.info('signal -%d received', signum)
        self.shutdown()

    def _startTasks(self):
        if self.REGISTER_SIGNAL_HANDLERS:
            # Things seem to fail more gracefully if we trigger the stop
            # out of band (with a signal handler) instead of catching the
            # KeyboardInterrupt...
            signal.signal(signal.SIGINT, self._handleShutdownSignals)
            signal.signal(signal.SIGTERM, self._handleShutdownSignals)
        for t in self.tasks:
            t.start()
        self.logger.debug("All tasks started")

    def getTask(self, name):
        """Returns a task for the given class `name` or type, or None."""
        for t in self.tasks:
            if isinstance(name, str):
                if t.name == name:
                    return t
            else:
                if t.__class__ is name:
                    return t
        return None

    def requireTask(self, name):
        """Returns a task for the given class `name` or type, or throws."""
        t = self.getTask(name)
        if t is None:
            raise Exception("Task %s not found in service" % name)
        return t

    def shutdown(self):
        """Request a graceful shutdown.  Does not block."""
        self.logger.info("Received graceful shutdown request")
        self.stop()

    def restart(self):
        """Request a graceful restart.  Does not block."""
        self.logger.info("Received graceful restart request")
        self._restart = True
        self.stop()

    def stop(self):
        self._stop = True

    def _wait(self):
        try:
            self.logger.debug('VService Active.  Awaiting graceful shutdown.')

            # If there are no remaining tasks (or this service has no tasks)
            # just sleep until ^C is pressed
            while not self._stop:
                time.sleep(0.1)
        except KeyboardInterrupt:
            self.logger.info('KeyboardInterrupt Received!  Stopping Tasks...')

        for t in reversed(self.tasks):
            t.stop()

        try:
            self.logger.info('Waiting for tasks to shutdown gracefully...')
            for t in reversed(self.tasks):
                self.logger.debug('Waiting for %s to stop...', t)
                t.join()
        except KeyboardInterrupt:
            self.logger.warning('Abandon all hope ye who enter here')

    def join(self):
        """Blocks until a stop is requested, waits for all tasks to shutdown"""
        while not self._stop:
            time.sleep(0.1)
        for t in reversed(self.tasks):
            t.join()

    @classmethod
    def initFromCLI(cls, name=None):
        """Starts this service, processing command line arguments."""
        ap = cls._buildArgumentParser()
        ns = ap.parse_args()
        instance = cls.initFromOptions(ns, name=name)
        return instance

    @classmethod
    def initFromOptions(cls, ns, name=None):
        """Starts this service, arguments from `ns`"""
        instance = cls(ns)
        if name is not None:
            instance.name = name
        instance.preprocessOptions()
        return cls._runloop(instance)

    @classmethod
    def _runloop(cls, instance):
        while not instance._stop:
            try:
                instance._createTasks()
                instance._startTasks()
            except Exception:
                instance.logger.exception("Unexpected Exception during init")
                instance.shutdown()

            instance._wait()

            if instance._restart:
                instance = cls(instance.options)

        instance.logger.info("Instance shut down gracefully")

    def startBG(self):
        """Starts this service in the background

        Returns a thread that will join() on graceful shutdown."""
        self._createTasks()
        self._startTasks()
        t = threading.Thread(target=self._wait)
        t.start()
        return t

    @property
    def name(self):
        if self._name is None:
            self._name = self.__class__.__name__
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def initLogging(self):
        """Basic stderr logging.  Override this to do something else."""
        logging.basicConfig(level=self.loglevel, stream=sys.stderr)

    @classmethod
    def _makeArgumentParser(cls):
        """Create an argparse.ArgumentParser instance.

        Override this method if you already have an ArgumentParser instance to use
        or you simply want to specify some of the optional arguments to
        argparse.ArgumentParser.__init__
        (e.g. "fromfile_prefix_chars" or "conflict_handler"...)
        """
        return ArgumentParser()

    @classmethod
    def _buildArgumentParser(cls):
        ap = cls._makeArgumentParser()
        cls._addArguments(ap)

        all_tasks = set(cls.TASKS).union(get_registered_tasks())
        for t in resolve_dependencies(all_tasks):
            # TODO: Add each tasks' arguments to an argument group
            t._addArguments(ap)
        return ap

    @property
    def loglevel(self):
        # TODO: Deprecate this after proting args to proper option()s
        return getattr(logging, self.options.level)

    def getOption(self, name, default=None):
        return getattr(self.options, name, default)

    def setOption(self, name, value):
        setattr(self.options, name, value)

    def getOptions(self):
        return self.options.__dict__

    def _install(self):
        if not HAS_PSUTIL:
            raise NotImplementedError("You need psutil installed to install "
                                      "under runit")
        import sparts.runit
        sparts.runit.install(self.name)
        sys.exit(0)

    def getChildren(self):
        return dict((t.name, t) for t in self.tasks)

    def getWarnings(self):
        return self.warnings

    def registerWarning(self, message):
        wid = self.warning_id
        self.warning_id += 1
        self.warnings[wid] = message
        return wid

    def clearWarnings(self):
        self.warnings = OrderedDict()

    def clearWarning(self, id):
        if id not in self.warnings:
            return False
        del self.warnings[id]
        return True



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Tools for dynamically generating thrift code"""

import distutils.spawn
import imp
import os.path
import tempfile

from six import iterkeys, iteritems
from sparts import ctx
from sparts.compat import OrderedDict, check_output
from sparts.fileutils import NamedTemporaryDirectory


def compile(path, root='.', debug=False, **kwargs):
    """Return a compiled thrift file module from `path`

    Additional kwargs may be passed to indicate options to the thrift compiler:

    - new_style [default:True]: Use new-style classes
    - twisted [default:False]: Generated twisted-friendly bindings
    - tornado [default:False]: Generate tornado-friendly bindings
    - utf8strings [default:False]: Use unicode strings instead of native
    - slots [default:True]: Use __slots__ in generated structs
    """
    comp = CompileContext(root=root, debug=debug)
    return comp.compileThriftFileAt(path, **kwargs)


def _require_executable(name):
    """Given `name`, assert on and return the path to that binary."""
    path = distutils.spawn.find_executable(name)
    assert path is not None, 'Unable to find %s in PATH' % repr(name)
    return path


class CompileContext(object):
    def __init__(self, root='.', debug=False):
        self.root = root
        self.thrift_bin = _require_executable('thrift')
        self.include_dirs = OrderedDict()
        self.dep_files = {}
        self.dep_contents = {}
        self.debug = debug

        self.addIncludeDir(self.root)

    def makeTemporaryIncludeDir(self):
        d = NamedTemporaryDirectory(prefix='tsrc_')
        if self.debug:
            d.keep()
        for k, v in iteritems(self.dep_contents):
            d.writefile(k, v)
        for k, v in iteritems(self.dep_files):
            d.symlink(k, v)
        return d

    def makeIncludeArgs(self, temp_include_dir=None):
        result = []
        for k in iterkeys(self.include_dirs):
            result += ['-I', k]

        if temp_include_dir is not None:
            result += ['-I', temp_include_dir.name]

        return result

    def getThriftOptions(self, new_style=True, twisted=False, tornado=False,
                         utf8strings=False, slots=True, dynamic=False,
                         dynbase=None, dynexc=None, dynimport=None):
        param = 'py'
        options = []
        if new_style:
            options.append('new_style')

        if twisted:
            options.append('twisted')
            assert not tornado

        if tornado:
            options.append('tornado')

        if utf8strings:
            options.append('utf8strings')

        if slots:
            options.append('slots')

        # TODO: Dynamic import jonx

        if len(options):
            param += ':' + ','.join(options)

        return param

    def addIncludeDir(self, path):
        assert os.path.exists(path) and os.path.isdir(path)
        self.include_dirs[os.path.abspath(path)] = True

    def addDependentFilePath(self, path):
        assert os.path.exists(path)

        self.dep_files[os.path.basename(path)] = os.path.abspath(path)

        path = os.path.dirname(path) or '.'
        self.addIncludeDir(path)

    def addDependentFileContents(self, name, contents):
        self.dep_contents[name] = contents

    def importThriftStr(self, payload, **kwargs):
        """Compiles a thrift file from string `payload`"""
        with tempfile.NamedTemporaryFile(suffix='.thrift') as f:
            if self.debug:
                f.delete = False
            f.write(payload)
            f.flush()
            return self.importThrift(f.name, **kwargs)

    def importThrift(self, path, **kwargs):
        """Compiles a .thrift file, importing its contents into its return value"""
        path = os.path.abspath(path)
        assert os.path.exists(path)
        assert os.path.isfile(path)

        srcdir = self.makeTemporaryIncludeDir()
        pathbase = os.path.basename(path)
        srcdir.symlink(pathbase, path)

        outdir = NamedTemporaryDirectory(prefix='to1_')
        outdir_recurse = NamedTemporaryDirectory(prefix='tor_')

        if self.debug:
            outdir.keep()
            outdir_recurse.keep()

        args = [self.thrift_bin] + self.makeIncludeArgs(srcdir) + \
               ["--gen", self.getThriftOptions(**kwargs), '-v',
                "-out", outdir.name, srcdir.join(pathbase)]
        check_output(args)

        args = [self.thrift_bin] + self.makeIncludeArgs(srcdir) + \
               ["--gen", self.getThriftOptions(**kwargs), '-v', '-r',
                "-out", outdir_recurse.name, srcdir.join(pathbase)]
        check_output(args)

        # Prepend output directory to the path
        with ctx.add_path(outdir_recurse.name, 0):

            thriftname = os.path.splitext(pathbase)[0]
            for dirpath, dirnames, filenames in os.walk(outdir.name):
                # Emulate relative imports badly
                dirpath = os.path.abspath(os.path.join(outdir, dirpath))
                with ctx.add_path(dirpath):
                    # Add types to module first
                    if 'ttypes.py' in filenames:
                        ttypes = self.importPython(dirpath + '/ttypes.py')
                        result = ttypes
                        filenames.remove('ttypes.py')

                    # Then constants
                    if 'constants.py' in filenames:
                        result = self.mergeModules(
                            self.importPython(dirpath + '/constants.py'),
                            result)
                        filenames.remove('constants.py')

                    for filename in filenames:
                        # Skip pyremotes
                        if not filename.endswith('.py') or \
                           filename == '__init__.py':
                            continue

                        # Attach services as attributes on the module.
                        svcpath = dirpath + '/' + filename
                        svcname = os.path.splitext(filename)[0]
                        svcmod = self.importPython(svcpath)
                        svcmod.__file__ = os.path.abspath(svcpath)
                        svcmod.__name__ = '%s.%s (generated)' % \
                            (thriftname, svcname)
                        setattr(result, svcname, svcmod)

            assert result is not None, "No files generated by %s" % (path, )

            # Set the __file__ attribute to the .thrift file instead
            # of the dynamically generated jonx
            result.__file__ = os.path.abspath(path)
            result.__name__ = thriftname + " (generated)"
            return result

    def mergeModules(self, module1, module2):
        if module1 is None:
            return module2

        if module2 is None:
            return module1

        for k in dir(module2):
            setattr(module1, k, getattr(module2, k))

        return module1

    def importPython(self, path):
        """Create a new module from code at `path`.

        Does not pollute python's module cache"""
        assert os.path.exists(path)

        # Any special variables we want to include in execution context
        orig_locals = {}
        exec_locals = orig_locals.copy()

        # Keep a copy of the module cache prior to execution
        with ctx.module_snapshot():
            execfile(path, exec_locals, exec_locals)

        # Generate a new module object, and assign the modified locals
        # as attributes on it.
        result = imp.new_module(path)
        for k, v in exec_locals.iteritems():
            setattr(result, k, v)

        return result



#
# Autogenerated by Thrift Compiler (1.0.0-dev)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import fb303.ttypes


from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None



class ServiceDefinition(object):
  """
  Attributes:
   - idl: The contents of the .thrift file used to generate this service
   - serviceName: A .thrift file can contain multiple services.  This field disambiguates
  between services returned in `idl`.
   - dependencies: Any additional (name).thrift => (contents) map of dependencies required
  to communicate with this service
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'idl', None, None, ), # 1
    (2, TType.STRING, 'serviceName', None, None, ), # 2
    (3, TType.MAP, 'dependencies', (TType.STRING,None,TType.STRING,None), None, ), # 3
  )

  def __init__(self, idl=None, serviceName=None, dependencies=None,):
    self.idl = idl
    self.serviceName = serviceName
    self.dependencies = dependencies

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.idl = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.serviceName = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.MAP:
          self.dependencies = {}
          (_ktype1, _vtype2, _size0 ) = iprot.readMapBegin()
          for _i4 in xrange(_size0):
            _key5 = iprot.readString();
            _val6 = iprot.readString();
            self.dependencies[_key5] = _val6
          iprot.readMapEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ServiceDefinition')
    if self.idl is not None:
      oprot.writeFieldBegin('idl', TType.STRING, 1)
      oprot.writeString(self.idl)
      oprot.writeFieldEnd()
    if self.serviceName is not None:
      oprot.writeFieldBegin('serviceName', TType.STRING, 2)
      oprot.writeString(self.serviceName)
      oprot.writeFieldEnd()
    if self.dependencies is not None:
      oprot.writeFieldBegin('dependencies', TType.MAP, 3)
      oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.dependencies))
      for kiter7,viter8 in self.dependencies.items():
        oprot.writeString(kiter7)
        oprot.writeString(viter8)
      oprot.writeMapEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Helpers for easily creating thrift clients"""
from __future__ import absolute_import

from thrift.transport.TSocket import TSocket
from thrift.transport.TTransport import TFramedTransport
from thrift.protocol.TBinaryProtocol import TBinaryProtocol

from functools import partial


class ThriftClient(object):
    """Base Class for easy interfacing with thrift services.

    `ThriftClient` can be used directly, or subclassed.  If subclassed,
    you can override a variety of attributes in order to make instantiation
    more natural:

        MODULE - The generated thrift module.  Should include Iface and Client
        HOST - A default host to connect to.  Possibly a vip.
        PORT - A default port to connect to.
        TRANSPORT_CLASS - By default, `TFramedTransport`
        PROTOCOL_CLASS - By default, `TBinaryProtocol`
        CONNECT_TIMEOUT - By default, 3.0 seconds

    Instantiation should be done via the class methods, `for_hostport()`, and
    `for_localhost()` as appropriate.  These helpers more aggressively require
    `port` and `host` arguments as appropriate.  Generic construction arguments
    override the class attribute defaults:

        `module` - Generated thrift module.
        `host` - IP address to connect to.
        `port` - Port to connect to.
        `connect_Timeout` - Socket connection timeout
        `transport_class` - Thrift transport class
        `protocol_class` - Thrift protocol class

    Additional features are configurable with other arguments:

        `lazy` - Default: True, connect on first RPC invocation, instead of
                 at construction time.

    Connections are made lazily, when the first rpc invocation occurs, so you
    do need to wrap client instantiation with a try-catch.
    """
    MODULE = None
    HOST = None
    PORT = None
    TRANSPORT_CLASS = TFramedTransport
    PROTOCOL_CLASS = TBinaryProtocol
    CONNECT_TIMEOUT = 3.0

    @classmethod
    def for_hostport(cls, host=None, port=None, **kwargs):
        assert host or cls.HOST, "You must define a host!"
        assert port or cls.PORT, "You must define a port!"
        return cls(host=host, port=port, **kwargs)

    @classmethod
    def for_localhost(cls, port=None, **kwargs):
        assert port or cls.PORT, "You must define a port!"
        return cls(host='127.0.0.1', port=port, **kwargs)

    def _initAttribute(self, name, value, default):
        if value is None:
            value = default
        setattr(self, name, value)

    def __init__(self, host=None, port=None, module=None, lazy=True,
                 connect_timeout=None, transport_class=None,
                 protocol_class=None):

        self._initAttribute('host', host, self.HOST)
        self._initAttribute('port', port, self.PORT)
        self._initAttribute('module', module, self.MODULE)
        self._initAttribute('connect_timeout', connect_timeout,
                            self.CONNECT_TIMEOUT)
        self._initAttribute('transport_class', transport_class,
                            self.TRANSPORT_CLASS)
        self._initAttribute('protocol_class', protocol_class,
                            self.PROTOCOL_CLASS)
        self.lazy = lazy

        assert self.module is not None, "You must define a thrift module!"

        if self.lazy:
            self._client = None
        else:
            self._connect()

    def _connect(self):
        # TODO: Add some kind of support for HTTP or SSLSocket
        self._socket = TSocket(self.host, self.port)
        self._socket.setTimeout(int(self.connect_timeout * 1000))
        self._transport = self.transport_class(self._socket)
        self._protocol = self.protocol_class(self._transport)
        self._client = self.module.Client(self._protocol)
        self._transport.open()

    def _lazyCall(self, name, *args, **kwargs):
        if self._client is None:
            self._connect()
        # TODO: Automatically connect on timed out connections
        return getattr(self._client, name)(*args, **kwargs)

    def __getattr__(self, name):
        getattr(self.module.Client, name)
        return partial(self._lazyCall, name)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
try:
    import unittest2 as unittest
except ImportError:
    import unittest
import logging
import os.path

from sparts.vservice import VService


# Import a Skip exception class that works with both pytest and unittest2
try:
    from _pytest.runner import Skipped
    class Skip(Skipped, unittest.SkipTest):
        pass

except ImportError:
    class Skip(unittest.SkipTest):
        pass


# Base test case for all sparts jonx
class BaseSpartsTestCase(unittest.TestCase):
    def assertNotNone(self, o, msg=''):
        self.assertTrue(o is not None, msg)

    def assertEmpty(self, arr, msg=''):
        return self.assertEquals(len(arr), 0, msg)

    def assertNotEmpty(self, o, msg=''):
        self.assertTrue(len(o) > 0, msg)

    def assertExists(self, path, msg=''):
        self.assertTrue(os.path.exists(path), msg)

    def assertNotExists(self, path, msg=''):
        self.assertFalse(os.path.exists(path), msg)

    @classmethod
    def setUpClass(cls):
        cls.logger = logging.getLogger('sparts.%s' % cls.__name__)
        super(BaseSpartsTestCase, cls).setUpClass()

    def setUp(self):
        if not hasattr(unittest.TestCase, 'setUpClass'):
            cls = self.__class__
            if not hasattr(cls, '_unittest2_setup'):
                cls.setUpClass()
                cls._unittest2_setup = 0
            cls._unittest2_setup += 1

    def tearDown(self):
        if not hasattr(unittest.TestCase, 'tearDownClass'):
            cls = self.__class__
            if not hasattr(cls, '_unittest2_setup'):
                cls._unittest2_setup = 0
            else:
                cls._unittest2_setup -= 1
            if cls._unittest2_setup == 0:
                cls.tearDownClass()

    def assertContains(self, item, arr, msg=''):
        return self.assertIn(item, arr, msg)

    @property
    def mock(self, *args, **kwargs):
        try:
            import mock
            return mock
        except ImportError:
            raise Skip("the mock module is required to run this test")


class ServiceTestCase(BaseSpartsTestCase):
    def getServiceClass(self):
        return VService

    def setUp(self):
        super(ServiceTestCase, self).setUp()

        TestService = self.getServiceClass()
        TestService.test = self

        ap = TestService._buildArgumentParser()
        ns = ap.parse_args(['--level', 'DEBUG'])
        self.service = TestService(ns)
        self.runloop = self.service.startBG()

    def tearDown(self):
        self.service.stop()
        self.runloop.join()
        super(ServiceTestCase, self).tearDown()


class MultiTaskTestCase(ServiceTestCase):
    TASKS = []

    def requireTask(self, task_name):
        self.assertNotNone(self.service)
        return self.service.requireTask(task_name)

    def getServiceClass(self):
        self.assertNotEmpty(self.TASKS)
        class TestService(VService):
            TASKS=self.TASKS
        return TestService

    def setUp(self):
        super(MultiTaskTestCase, self).setUp()
        for t in self.TASKS:
            self.service.requireTask(t.__name__)


class SingleTaskTestCase(MultiTaskTestCase):
    TASK = None

    @classmethod
    def setUpClass(cls):
        super(SingleTaskTestCase, cls).setUpClass()
        if cls.TASK:
            cls.TASKS = [cls.TASK]

    def setUp(self):
        self.assertNotNone(self.TASK)
        super(SingleTaskTestCase, self).setUp()
        self.task = self.service.requireTask(self.TASK.__name__)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Run-time compatiblity helpers with third-party modules

For most standard library moves and discrepancies, you should use `six` instead
"""
try:
    from collections import OrderedDict
except ImportError:
    # Python2.6 compatibility
    from ordereddict import OrderedDict

try:
    from subprocess import check_output
except ImportError:
    import subprocess

    def check_output(args, stdin=None, stderr=None, shell=False,
                     universal_newlines=False):
        """Mostly compatible `check_output` for python2.6"""
        p = subprocess.Popen(args, stdin=stdin, stderr=stderr, shell=shell,
              universal_newlines=universal_newlines)
        output, stderr = p.communicate()
        returncode = p.wait()
        if returncode != 0:
            raise subprocess.CalledProcessError(returncode, cmd=args, output=output)
        return output



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
from .periodic import PeriodicTask
from threading import Event


class PollerTask(PeriodicTask):
    """A PeriodicTask oriented around monitoring a single value.
    
    Simply override `fetch`, and the `onValueChanged()` method will be called
    with the old and new values.  Additionally, the `getValue()` method can
    be called by other tasks to block until the values are ready.
    """
    def initTask(self):
        self.current_value = None
        self.fetched = Event()
        super(PollerTask, self).initTask()

    def execute(self, context=None):
        new_value = self.fetch()
        if self.current_value != new_value:
            self.onValueChanged(self.current_value, new_value)
        self.current_value = new_value
        self.fetched.set()

    def onValueChanged(self, old_value, new_value):
        self.logger.debug('onValueChanged(%s, %s)', old_value, new_value)

    def fetch(self):
        self.logger.debug('fetch')
        return None

    def getValue(self, timeout=None):
        self.fetched.wait(timeout)
        return self.current_value



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""thrift-related helper tasks"""
from __future__ import absolute_import

from ..vtask import VTask

from sparts.sparts import option
from thrift.server.TNonblockingServer import TNonblockingServer
from thrift.transport.TSocket import TServerSocket

import time


class ThriftProcessorTask(VTask):
    """A loopless task that helps process thrift requests.

    You will need a subclass of this task in order to properly handle thrift
    requests in your NBServerTasks and such."""
    LOOPLESS = True
    PROCESSOR = None

    def __init__(self, service):
        # TODO: I don't like the way this works: it's too much boilerplate and
        # complexity to implement custom thrift services.
        super(ThriftProcessorTask, self).__init__(service)
        assert self.PROCESSOR is not None
        self.processor = self.PROCESSOR(self.service)


class NBServerTask(VTask):
    """Spin up a thrift TNonblockingServer in a sparts worker thread"""
    DEFAULT_HOST = '0.0.0.0'
    DEFAULT_PORT = 0
    OPT_PREFIX = 'thrift'

    bound_host = bound_port = None

    host = option(default=lambda cls: cls.DEFAULT_HOST, metavar='HOST',
                  help='Address to bind server to [%(default)s]')
    port = option(default=lambda cls: cls.DEFAULT_PORT,
                  type=int, metavar='PORT',
                  help='Port to run server on [%(default)s]')
    num_threads = option(name='threads', default=10, type=int, metavar='N',
                         help='Server Worker Threads [%(default)s]')

    def getProcessor(self):
        """Automatically find the ThriftProcessorTask subclass"""
        found = None
        for task in self.service.tasks:
            if isinstance(task, ThriftProcessorTask):
                assert found is None, "Multiple processor tasks! (%s, %s)" % \
                    (found.name, task.name)
                found = task
        assert found is not None, "No ThriftProcessorTask's found!"
        return found.processor

    def initTask(self):
        """Overridden to bind sockets, etc"""
        super(NBServerTask, self).initTask()

        self._stopped = False
        self.socket = TServerSocket(self.host, self.port)
        self.server = TNonblockingServer(self.getProcessor(), self.socket,
                                         threads=self.num_threads)
        self.server.prepare()
        self.bound_host, self.bound_port = \
            self.server.socket.handle.getsockname()
        self.logger.info("%s Server Started on %s:%s",
                         self.name, self.bound_host, self.bound_port)

    def stop(self):
        """Overridden to tell the thrift server to shutdown asynchronously"""
        self.server.stop()
        self.server.close()
        self._stopped = True

    def _runloop(self):
        """Overridden to execute TNonblockingServer's main loop"""
        while not self.server._stop:
            self.server.serve()
        while not self._stopped:
            time.sleep(0.1)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Module related to configuring services under runit"""
import sys
import os.path
from .fileutils import writefile, find_executable, resolve_partition, makedirs
import logging
import stat

logger = logging.getLogger('sparts.runit')

def install(service_name):
    """Installs the running python script as `service_name` under runit."""
    preferred = '/etc/service'
    dirs = get_runsvdir_dirs()
    assert len(dirs) > 0, "runsvdir is not running!"

    if preferred not in dirs:
        preferred = dirs[0]

    service_path = os.path.join(preferred, service_name)
    logger.info('Installing %s in %s', service_name, service_path)
    make_runit_dir(service_name, service_path)

def is_runit_installed():
    """Returns True if runit is installed"""
    # Check if the `runsv` binary is in the path.
    return bool(find_executable('runsv'))

def get_runsvdir_dirs():
    """Returns all dirs being currently managed by `runsvdir`"""
    import psutil
    dirs = []
    # Find all running `runsvdir` processes
    for proc in psutil.process_iter():
        if proc.name == 'runsvdir':
            d = get_runsvdir_dir_from_cmdline(proc.cmdline)
            if d is not None:
                dirs.append(d)
    return dirs

def get_runsvdir_dir_from_cmdline(cmdline):
    """Return runsvdir's target path based on its `cmdline` args"""
    # TODO - unittest this
    for i, arg in enumerate(cmdline):
        # Skip the process name
        if i == 0:
            continue

        # Skip the -P, -H flags
        if arg in ['-P', '-H']:
            continue

        return arg
    return None

def on_same_filesystem(path1, path2):
    """Returns True if `path` and `path2` reside on the same mount"""
    return resolve_partition(path1).mountpoint == \
            resolve_partition(path2).mountpoint

def get_default_args():
    args = sys.argv[:]
    assert len(args) > 0, "Something went horribly wrong"
    assert sys.executable is not None, "Something went horribly wrong"
    args.insert(0, sys.executable)

    # runsv will only redirect standard output to logs.
    # sparts logs stderr only by default.  Let's grab everything
    args.append('2>&1')
    return args

def make_runit_dir(service_name, path, args=None, make_logdir=True):
    # TODO - unittest this
    if args is None:
        args = get_default_args()

    makedirs(path)

    if make_logdir:
        svlogd = find_executable('svlogd')
        assert svlogd is not None, "Unable to make runit dir without svlogd"

        logdir = os.path.join('/var/log', service_name)
        makedirs(logdir)
        make_runit_dir(
            service_name + '.log', os.path.join(path, 'log'),
            args=[svlogd, '-ttv', logdir],
            make_logdir=False
        )

    run_path = os.path.join(path, 'run')
    writefile(run_path, make_run_script_for_args(args))
    flags = os.stat(run_path).st_mode
    os.chmod(run_path, flags | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


def make_run_script_for_args(args):
    # TODO - unittest this
    parts = []
    for arg in args:
        if arg == '--runit-install':
            continue

        if os.path.exists(arg):
            parts.append(os.path.realpath(arg))
        else:
            parts.append(arg)

    return "#!/bin/bash\n" + \
        'exec ' + ' '.join(parts)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
from setuptools import setup, find_packages, Command
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.test import test as TestCommand

from distutils.spawn import find_executable
from glob import glob
import os.path
import imp
import sys


THRIFT = find_executable('thrift')

ROOT = os.path.abspath(os.path.dirname(__file__))

def read(fname):
    """Read a file relative to the repository root"""
    return open(os.path.join(ROOT, fname)).read()

def exists(fname):
    """Returns True if `fname` relative to `ROOT` exists"""
    return os.path.exists(os.path.join(ROOT, fname))

def version():
    """Return the version number from sparts/__version__.py"""
    file, pathname, description = imp.find_module('sparts', [ROOT])
    return imp.load_module('sparts', file, pathname, description).__version__

# Initialize custom command handlers
cmdclass = {}

# These files are shadowed in the source repository from
# externals.  If you are developing sparts, you can use git submodule to make
# sure you have the latest/greatest fb303 from thrift.
WANT_COPY = {
    'externals/thrift/contrib/fb303/if/fb303.thrift':
        'thrift/fb303.thrift',
}

# Let's figure out which files exist in which submodules...
CAN_COPY = []
for src in WANT_COPY:
    if exists(src):
        CAN_COPY.append(src)


class submodule_copy(Command):
    user_options=[]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for src in CAN_COPY:
            self.copy_file(os.path.join(ROOT, src),
                           os.path.join(ROOT, WANT_COPY[src]))

if CAN_COPY:
    cmdclass['submodule_copy'] = submodule_copy

# If we have a thrift compiler installed, let's use it to re-generate
# the .py files.  If not, we'll use the pre-generated ones.
if THRIFT is not None:
    class gen_thrift(Command):
        user_options=[]

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            self.mkpath(os.path.join(ROOT, 'sparts', 'gen'))
            for f in glob(os.path.join(ROOT, 'thrift', '*.thrift')):
                self.spawn([THRIFT, '-out', os.path.join(ROOT, 'sparts', 'gen'),
                            '-v', '--gen', 'py:new_style',
                            os.path.join(ROOT, 'thrift', f)])

    cmdclass['gen_thrift'] = gen_thrift


# Custom build_py handler.  Triggers submodule_copy and gen_thrift
# if the environment is right.
class build_py(_build_py):
    def run(self):
        if CAN_COPY:
            self.run_command('submodule_copy')

        if THRIFT is not None:
            self.run_command('gen_thrift')
        _build_py.run(self)

cmdclass['build_py'] = build_py


# Custom PyTest Test command, per https://pytest.org/latest/goodpractises.html
class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests', '-rfEsx']
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

cmdclass['test'] = PyTest

install_requires = ['six']
if sys.version < '2.7':
    install_requires.append('ordereddict')

tests_require = install_requires + ['pytest', 'mock']
if sys.version < '2.7':
    tests_require.append('unittest2')

VERSION = version()
setup(
    name="sparts",
    version=VERSION,
    packages=find_packages(exclude=['tests', 'tests.*']),
    description="Build services in python with as little code as possible",
    long_description=read("README.rst"),

    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'thrift': ['thrift'],
        'tornado': ['tornado'],
        'twisted': ['Twisted'],
    },
    author='Peter Ruibal',
    author_email='ruibalp@gmail.com',
    license='ISC',
    keywords='service boostrap daemon thrift tornado',
    url='http://github.com/facebook/sparts',
    download_url='https://github.com/facebook/sparts/archive/%s.tar.gz' % VERSION,

    test_suite="tests",
    cmdclass=cmdclass,

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
    ],
)



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
import imp

def HAS(module):
    try:
        file, pathname, description = imp.find_module(module)
        return imp.load_module(module, file, pathname, description)
    except ImportError:
        return None

HAS_PSUTIL = HAS('psutil')
HAS_THRIFT = HAS('thrift')



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Twisted-related helper tasks"""
from __future__ import absolute_import

from ..vtask import VTask, SkipTask

from twisted.internet import epollreactor
epollreactor.install()

import twisted.internet


class TwistedReactorTask(VTask):
    """Configure and run the twisted reactor in a sparts task"""
    reactor = None

    def initTask(self):
        super(TwistedReactorTask, self).initTask()
        needed = getattr(self.service, 'REQUIRE_TWISTED', False)
        for t in self.service.tasks:
            if isinstance(t, TwistedTask):
                needed = True

        if not needed:
            raise SkipTask("No TwistedTasks found or enabled")

        self.reactor = twisted.internet.reactor

    def start(self):
        # TODO: register signals manually using some 'clean' signal handler
        # chaining stuff
        self.reactor._handleSignals()
        super(TwistedReactorTask, self).start()

    def _runloop(self):
        self.reactor.run(installSignalHandlers=0)

    def stop(self):
        self._tryShutdown()

    def _tryShutdown(self):
        can_shutdown = True
        self.logger.debug('_tryShutdown with %d tasks', len(self.service.tasks))
        for t in self.service.tasks:
            if isinstance(t, TwistedTask):
                if not t.isDoneWithReactor():
                    self.logger.debug("%s is not done with the reactor", t.name)
                    can_shutdown = False
        if can_shutdown:
            self.reactor.callFromThread(self.reactor.stop)
        else:
            self.reactor.callFromThread(self.reactor.callLater, 0.3,
                                        self._tryShutdown)

class TwistedTask(VTask):
    """Base class for tasks that require a twisted reactor.
    
    Implicitly configures the twisted reactor task as a dependency."""
    DEPS = [TwistedReactorTask]

    def initTask(self):
        super(TwistedTask, self).initTask()
        self.reactor_task = self.service.requireTask('TwistedReactorTask')

    @property
    def reactor(self):
        return self.reactor_task.reactor

    def isDoneWithReactor(self):
        """Override this to handle shutdowns more gracefully"""
        return True



# Copyright (c) 2014, Facebook, Inc.  All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree. An additional grant
# of patent rights can be found in the PATENTS file in the same directory.
#
"""Tasks related for files and filesystems"""
from .poller import PollerTask
import errno
import os

from ..sparts import option

class DirectoryWatcherTask(PollerTask):
    """DirectoryWatcherTask watches for new, deleted, and modified files
    
    The `IGNORE_INITIAL_FILES` attribute can be overridden to Flaase if you
    do not want to receive a bunch of `onFileCreated` callbacks during startup.

    This could be better implemented with the inotify/pyinotify API, but for
    basic rapid prototyping, polling should be sufficient.
    """
    PATH = '.'
    path = option(default=lambda cls: cls.PATH,
                  help='Directory path to watch [%(default)s]')
    IGNORE_INITIAL_FILES = True

    def onFileCreated(self, filename, stat):
        """Override this to do custom processing when new files are created."""
        self.logger.debug('onFileCreated(%s, %s)', filename, stat)

    def onFileDeleted(self, filename, old_stat):
        """Override this to do custom processing when files are deleted."""
        self.logger.debug('onFileDeleted(%s, %s)', filename, old_stat)

    def onFileChanged(self, filename, old_stat, new_stat):
        """Override this to do custom processing when files are modified."""
        self.logger.debug('onFileChanged(%s, %s, %s)', filename, old_stat, new_stat)


    def fetch(self):
        """Overridden to stat a particular filesystem path"""
        d = {}
        root = self.path
        contents = self.listdir(root)
        for name in contents:
            try:
                d[name] = self.stat(os.path.join(root, name))
            except OSError as e:
                # There is a race condition between listdir() and stat() where
                # files might be deleted.  On ENOENT from a stat call, assume
                # the file is gone.  Raise non-ENOENT exceptions
                if e.errno != errno.ENOENT:
                    raise

        return sorted(d.items())

    def onValueChanged(self, old_value, new_value):
        """Overridden to track file statuses."""
        if old_value is None:
            # For the first run, ignore pre-existing files by default.
            # Override this class and set this to False if you want them.
            if self.IGNORE_INITIAL_FILES:
                return

            old_value = []
        old_value_dict = dict(old_value)
        new_value_dict = dict(new_value)

        for (name, old_stat) in old_value:
            if name not in new_value_dict:
                self.onFileDeleted(name, old_stat)
            else:
                new_stat = new_value_dict[name]
                if new_stat != old_stat:
                    self.onFileChanged(name, old_stat, new_stat)

        for (name, new_stat) in new_value:
            if name not in old_value_dict:
                self.onFileCreated(name, new_stat)

    def listdir(self, path):
        """Wrapper for making unittesting/mocking easier"""
        return os.listdir(path)

    def stat(self, path):
        """Wrapper for making unittesting/mocking easier"""
        return os.stat(path)
